/* Copyright (c) 2000-2002, Jacco Bikker
 * Copyright (c) 2002-2003, Gary Dubuque
 * Copyright (c) 2003-2004, Kino Coursey (Mods for CyN)
 *
 * This file is part of NP+AIML.
 * NP+AIML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation; either version 2, or (at your option) any later 
 * version.
 *
 * NP+AIML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.

 * You should have received a copy of the GNU General Public License along 
 * with NP+AIML; see the file COPYING. If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

// AIMLpadView.h : interface of the CAIMLpadView class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_AIMLPADVIEW_H__84EF1BB8_9AAB_4DCB_A477_05CF3FB934CC__INCLUDED_)
#define AFX_AIMLPADVIEW_H__84EF1BB8_9AAB_4DCB_A477_05CF3FB934CC__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <afxdlgs.h>
#include <afxwin.h>

#include "LearnDlg.h"
#include "CmdDlg.h"
#include "SubjectsDlg.h"
#include "CasesDlg.h"
#include "FactsDlg.h"
#include "TreeDlg.h"
#include "ScriptDlg.h"
#include "FrameDlg.h"
#include "AttrDlg.h"
#include "RuleDlg.h"
#include "DumpDlg.h"
#include "ListDlg.h"
#include "TargetDlg.h"
#include "SearchDlg.h"

#include "gotodlg.h"

#include "core.h"
#include "pu.h"

class CAttrDlg;
class CRuleDlg;

class CAIMLpadView : public CEditView
{
protected: // create from serialization only
	CAIMLpadView();
	DECLARE_DYNCREATE(CAIMLpadView)

// Attributes
public:
	CAIMLpadDoc* GetDocument();

	CLearnDlg*		m_pLearnDialog;			// load AIML dialog
	CCmdDlg*		m_pCmdDialog;			// command dialog

	CSubjectsDlg*	m_pSubjectDialog;		// subjects dialog
	CCasesDlg*		m_pCaseDialog;			// cases dialog
	CFactsDlg*		m_pFactDialog;			// facts dialog
	CTreeDlg*		m_pTreeDialog;			// knowledge tree dialog

	CScriptDlg*		m_pScriptDialog;		// scripts tree dialog
	CFrameDlg*		m_pFrameDialog;			// frame dialog

	CAttrDlg*	 m_pAttrDialog;				// attribute dialog
	CRuleDlg*	 m_pRuleDialog;				// rule dialog

	CGotoDlg*		m_pGotoDialog;          // goto line number dialog

	CDumpDlg*		m_pDumpDialog;
	CListDlg*		m_pListDialog;
	CSearchDlg*		m_pSearchDialog;
	CTargetDlg*		m_pTargetDialog;

	Core bot;
	PUvm pu;
	bool blnThinking;

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAIMLpadView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnInitialUpdate();
	protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CAIMLpadView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

private:
	bool blnDialog;
	bool blnTarget;
	bool blnUseMSAgent;
	bool blnMSAgent;
	bool blnMSAgentActive;

    bool blnGraphmaster;

	int atpos;
	int topos;
	int atline;

// Generated message map functions
protected:
	//{{AFX_MSG(CAIMLpadView)
	afx_msg void OnAimlLearn();
	afx_msg void OnAimlSort();
	afx_msg void OnAimlDump();
	afx_msg void OnAimlRemoveDups();
	afx_msg void OnAimlReindex();
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnDestroy();
	afx_msg void OnDialog();
	afx_msg void OnTarget();
	afx_msg void OnCommand();
	afx_msg void OnCases();
	afx_msg void OnFacts();
	afx_msg void OnSubjects();
	afx_msg void OnTree();
	afx_msg void OnFrame();
	afx_msg void OnScripts();
	afx_msg void OnEditGotoline();
	afx_msg void OnToolsAimlEditTargets();
	afx_msg void OnToolsAimlQuickCategory();
	afx_msg void OnToolsAimlClassify();
	afx_msg void OnToolsAimlList();
	afx_msg void OnToolsContextSetgossip();
	afx_msg void OnToolsContextSettarget();
	afx_msg void OnToolsAimlLoadtargets();
	afx_msg void OnToolsAimlSearchCategories();
	afx_msg void OnViewUseMsagent();
	afx_msg void OnUpdateToolsContextUseMsagent(CCmdUI* pCmdUI);
	afx_msg void OnAttributes();
	afx_msg void OnRules();
	afx_msg void OnToolsUsegraphmaster();

	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  // debug version in AIMLpadView.cpp
inline CAIMLpadDoc* CAIMLpadView::GetDocument()
   { return (CAIMLpadDoc*)m_pDocument; }
#endif

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_AIMLPADVIEW_H__84EF1BB8_9AAB_4DCB_A477_05CF3FB934CC__INCLUDED_)
