/* Copyright (c) 2000-2002, Jacco Bikker
 * Copyright (c) 2002-2003, Gary Dubuque
 *
 * This file is part of NP+AIML.
 * NP+AIML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation; either version 2, or (at your option) any later 
 * version.
 *
 * NP+AIML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.

 * You should have received a copy of the GNU General Public License along 
 * with NP+AIML; see the file COPYING. If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

// CasesDlg.cpp : implementation file
//

#include "stdafx.h"
#include "AIMLpad.h"
#include "CasesDlg.h"
#include "FactsDlg.h"
#include "pu.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCasesDlg dialog


CCasesDlg::CCasesDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CCasesDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CCasesDlg)
	m_case = _T("");
	m_subject = _T("");
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CCasesDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CCasesDlg)
	DDX_Control(pDX, IDC_CASELIST, m_cases);
	DDX_Text(pDX, IDC_CASENAME, m_case);
	DDX_Text(pDX, IDC_SUBJECTSTATUS, m_subject);
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


void CCasesDlg::FillList()
{
	CString key;
	Case *pCase;
	Subject *pSubj = kb->curSub;

	m_cases.ResetContent();

	POSITION pos = pSubj->m_cases.m_mapStringToCase.GetStartPosition();
	while (pos != NULL) {
		pSubj->m_cases.m_mapStringToCase.GetNextAssoc(pos,key,pCase);
		m_cases.AddString(pCase->m_name);
	}

	UpdateData(false);
}


BOOL CCasesDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	FillList();

	CWnd* pWnd = GetDlgItem(IDC_CASENAME);
	pWnd->SetFocus();

	return false;
}


BEGIN_MESSAGE_MAP(CCasesDlg, CDialog)
	//{{AFX_MSG_MAP(CCasesDlg)
	ON_BN_CLICKED(IDC_MAKE, OnMake)
	ON_BN_CLICKED(IDC_DROP, OnDrop)
	ON_LBN_DBLCLK(IDC_CASELIST, OnDblclkCaselist)
	ON_BN_CLICKED(IDC_LOAD_CASE, OnLoadCase)
	ON_BN_CLICKED(IDC_SAVE_CASE, OnSaveCase)
	ON_BN_CLICKED(IDC_LINK_FACTS, OnLinkFacts)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCasesDlg message handlers

void CCasesDlg::OnLoadCase() 
{
	char *fn = new char[MAX_PATH_LEN];
	CString szFilter = "Text Files (*.txt)|*.txt|All Files (*.*)|*.*||";
	CFileDialog dlg(true, _T("txt"), NULL, OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY, szFilter, NULL);

	dlg.m_ofn.lpstrTitle = TEXT("Load Case From File...");

	if (dlg.DoModal() != IDCANCEL)	{
		GetParent()->BeginWaitCursor();
		sprintf(fn,"%s",m_case);
		pu->utmake(fn);
		sprintf(fn,"\"%s\"",dlg.GetPathName());
		pu->utload(fn);
		GetParent()->EndWaitCursor();
	}
	delete dlg;
	delete fn;

	FillList();
}

void CCasesDlg::OnSaveCase() 
{
	char *fn = new char[MAX_PATH_LEN];
	CString szFilter = "Text Files (*.txt)|*.txt|All Files (*.*)|*.*||";
	CFileDialog dlg(false, _T("txt"), NULL, OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY, szFilter, NULL);

	dlg.m_ofn.lpstrTitle = TEXT("Save Case To File...");

	if (dlg.DoModal() != IDCANCEL)	{
		GetParent()->BeginWaitCursor();
		sprintf(fn,"%s",m_case);
		pu->utmake(fn);
		sprintf(fn,"\"%s\"",dlg.GetPathName());
		pu->utsave(fn);
		GetParent()->EndWaitCursor();
	}

	delete dlg;
	delete fn;
}

void CCasesDlg::OnMake() 
{
	char *buf;

	UpdateData(true);

	buf = new char[m_case.GetLength() + 2];
	sprintf(buf,"%s",m_case);
	if (kb->curSub->m_cases.Find(buf)) {
		kb->curTag->m_cases.Find(buf);
	} else {
		if (MessageBox(_T("Ok to add new case?"),m_case,MB_YESNO) == IDYES) {
			kb->curSub->m_cases.Add(buf);
			kb->curTag->m_cases.Add(buf);
		}
	}
	delete buf;

	FillList();
}

void CCasesDlg::OnDrop() 
{
	char *buf;

	UpdateData(true);

	buf = new char[m_case.GetLength() + 2];
	sprintf(buf,"%s",m_case);

	if (strcmp(buf,"~~global~~")) {
		if (kb->curSub->m_cases.Find(buf)) {
			if (MessageBox(_T("Ok to delete case and all facts and events related to it?"),m_case,MB_YESNO) == IDYES) {
				m_case = _T("");
				if (kb->curSub->m_cases.Remove(buf)) {
					kb->curSub->m_cases.curCase = kb->curSub->m_cases.curGbl;
					m_case = kb->curSub->m_cases.curCase->m_name;
				}
				if (kb->curTag->m_cases.Remove(buf)) {
					kb->curTag->m_cases.curCase = kb->curTag->m_cases.curGbl;
				}
			}
		} else m_case = _T("");
	} else if (MessageBox(_T("Ok to clear global case of all facts?"),m_case,MB_YESNO) == IDYES) kb->curSub->m_cases.Add(buf);

	delete buf;

	FillList();
}

void CCasesDlg::OnDblclkCaselist() 
{
	int nIndex = m_cases.GetCurSel();
	if (nIndex != LB_ERR) {
		m_cases.GetText(nIndex,m_case);
		UpdateData(false);
	}
}

void CCasesDlg::OnLinkFacts() 
{
	char *buf;
	CString str;
	CFactsDlg *m_pFactDialog = new CFactsDlg;
	Case *pCase = kb->curSub->m_cases.curCase;

	UpdateData(true);

	buf = new char[m_case.GetLength() + 2];
	sprintf(buf,"%s",m_case);
	if (kb->curSub->m_cases.Find(buf)) {
		m_pFactDialog->kb = &(pu->kb);

		str.Format(_T("%s.%s"),pu->kb.curSub->m_name,pu->kb.curSub->m_cases.curCase->m_name);

		m_pFactDialog->m_case = str;
		m_pFactDialog->m_fact = _T("");
		
		m_pFactDialog->DoModal();
	}
	kb->curSub->m_cases.curCase = pCase;

	delete m_pFactDialog;
	delete buf;
}
