// Code to use Microsoft Agent v2.0 characters with AIMLpad.
// This code was massaged from the "Hello.cpp" example code that
// Microsoft uses to demonstrate how to use MSAgent characters.
// Notes:
// You MUST have/use the MSAgent v2.0 include and IDL files.
// Stefan Zakarias, 19-MAR-2003.
//
// Additional mods made for CyN (2003-2004)
#include "stdafx.h"

#include <windows.h>
#include <tchar.h>
#include "AgtSvr.h"
#include "AgtSvr_i.c"


static const LPWSTR kpwzCharacterOld = L"\\program files\\microsoft agent\\characters\\genie.acs";
static const LPTSTR kpszAppTitle = _T("AIMLpad - MSAgent");
static const LPWSTR kpswLHVoice = L"{227A0E40-A92A-11d1-B17B-0020AFED142E}";

static long lCharID;
static long lRequestID;
static VARIANT vMSAgentPath;
static VARIANT vTTSModeID;
static IAgentEx *pAgentEx;
static IAgentCharacterEx *pCharacterEx;

void 
ShutdownMSAgent()
{
	// Clean up
	if (pCharacterEx) 
	{
		// We release the character interface.
		pCharacterEx->Release();

		// Just releasing the character interface does NOT make
		// the character go away.  You've got to call Unload() too!
		pAgentEx->Unload(lCharID);
	}
	
	// Release the Agent
	if (pAgentEx) 
	{
		pAgentEx->Release();
	}
	// Clear the path to the MSAgent 'xxx.acs' file.
	VariantClear(&vMSAgentPath);

	if (pCharacterEx)
	{
		// Free the COM resources.
		CoUninitialize();

		// Make sure we don't try to free things twice.
		pAgentEx = NULL;
		pCharacterEx = NULL;
	}
}

bool
StartMSAgent(char *pMSAgentFile)
{
	HRESULT hRes;
	_TCHAR szError[256];

	pAgentEx = NULL;
	pCharacterEx = NULL;
	
	// Initialize the COM library
    if (FAILED(CoInitialize(NULL))) 
	{
		MessageBox(NULL, _T("Error initializing COM."), kpszAppTitle, MB_OK | MB_ICONERROR);
        return false;
	}

	// If MSAgent server is running, this function connects to the server;
	// Otherwise we'll create an instance of the MSAgent 2.0 server and start it.
	// By asking for an IAgentEx interface we will get at least MSAgent 2.0.
	hRes = CoCreateInstance(CLSID_AgentServer, NULL, CLSCTX_SERVER, IID_IAgentEx, (LPVOID *)&pAgentEx);

	if (FAILED(hRes)) 
	{
		wsprintf(szError, _T("Error initializing MSAgent, code = 0x%x"), hRes);
		MessageBox(NULL, szError, kpszAppTitle, MB_OK | MB_ICONERROR | MB_TOPMOST);
		CoUninitialize();

		return false;
	}

	__try 
	{
		// First try to load the user's desired character
		if (pMSAgentFile)
		{
			// Buffer for the converted string
			unsigned short ppszW[1024];
			ULONG cCharacters = 0;

			// Get the string length.
			cCharacters = strlen(pMSAgentFile)+1;

			// Convert the string 'pszA' to a wide string and place it in 'ppszW[]'.
			if (MultiByteToWideChar(CP_ACP, 0, pMSAgentFile, cCharacters, ppszW, 1024))
			{
				BSTR bszMSagentPath = SysAllocString(ppszW);

				if (bszMSagentPath == NULL) 
				{
					hRes = E_OUTOFMEMORY;
					__leave;
				}

				vMSAgentPath.vt = VT_BSTR;
				vMSAgentPath.bstrVal = bszMSagentPath;

				hRes = pAgentEx->Load(vMSAgentPath, &lCharID, &lRequestID);


				SysFreeString(bszMSagentPath);
			}
		}
		else
		{
			// No user specified character...
			// Set flag to "FAILED" so that we can try below.
			hRes = -1;
		}	
		
		// Have we successfully loaded a character?
		if (FAILED(hRes))
		{
			// If there's no user specified character then try to load the default one.
			vMSAgentPath.vt = VT_EMPTY;

			hRes = pAgentEx->Load(vMSAgentPath, &lCharID, &lRequestID);

			if (FAILED(hRes)) 
			{
				// There's no default character. See if we can load the
				// character from the directory used in most version 1.5
				// applications.

				BSTR bszMSagentPath = SysAllocString(kpwzCharacterOld);
				
				if (bszMSagentPath== NULL) 
				{
					hRes = E_OUTOFMEMORY;
					__leave;
				}
	
				vMSAgentPath.vt = VT_BSTR;
				vMSAgentPath.bstrVal = bszMSagentPath;

				hRes = pAgentEx->Load(vMSAgentPath, &lCharID, &lRequestID);

				SysFreeString(bszMSagentPath);

				// Did we successfully load a character?
				if (FAILED(hRes))
				{
					__leave;
				}
			}
		}

		// Get the IAgentCharacterEx interface
		hRes = pAgentEx->GetCharacterEx(lCharID, &pCharacterEx);

		if (FAILED(hRes))
			__leave;

		// Set the language of the MSAgent character
		hRes = pCharacterEx->SetLanguageID(MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US));

		if (FAILED(hRes))
			__leave;


		// Yay!  A successful exit!
		return true;
	}

	__finally 
	{
		if (FAILED(hRes)) 
		{
			// If result is not equal to "cannot find MSAgent character(s)"
			if (hRes != 0x80070003)
			{
				wsprintf(szError, _T("Error occurred in the MSAgent, code = 0x%x"), hRes);
				MessageBox(NULL, szError, kpszAppTitle, MB_OK | MB_ICONERROR | MB_TOPMOST);
			}
		}
	}

	// If we get to here, then something didn't work.
	ShutdownMSAgent();

	// Return "false" for failure.
	return false;
}

HRESULT
ShowMSAgent()
{
	// Show the character.  The first parameter tells Microsoft
	// Agent to show the character by playing an animation.
	return(pCharacterEx->Show(FALSE, &lRequestID));
}

HRESULT
HideMSAgent()
{
	return(pCharacterEx->Hide(FALSE, &lRequestID));
}

void
SpeakMSAgent(LPCSTR pszA)
{
	// We must first convert the ANSI string 'pszA' to a Unicode string.

	// Buffer for the converted string
	unsigned short ppszW[1024];
    ULONG cCharacters;

    // If input is null then just return.
    if (NULL == pszA)
        return;

    // Get the string length.
    cCharacters =  strlen(pszA)+1;

    // Covert to string 'pszA' to Unicode and place it in 'ppszW[]'.
    if (0 == MultiByteToWideChar(CP_ACP, 0, pszA, cCharacters, ppszW, 1024))
    {
		// If we got screwed then just return.
        return;
    }

	// Now we can make the MSAgent character speak to us.
	BSTR bszSpeak = SysAllocString(ppszW);
	pCharacterEx->Speak(bszSpeak, NULL, &lRequestID);
	SysFreeString(bszSpeak);
}

void
SetMSAgentTTS(LPCSTR pszA)
{
	// We must first convert the ANSI string 'pszA' to a Unicode string.
	HRESULT hRes;

	// Buffer for the converted string
	unsigned short ppszW[1024];
    ULONG cCharacters;

    // If input is null then just return.
    if (NULL == pszA)
        return;

    // Get the string length.
    cCharacters =  strlen(pszA)+1;

    // Covert to string 'pszA' to Unicode and place it in 'ppszW[]'.
    if (0 == MultiByteToWideChar(CP_ACP, 0, pszA, cCharacters, ppszW, 1024))
    {
		// If we got screwed then just return.
        return;
    }

				//Character.LanguageID = &H409
				//Character.TTSModeID = "{227A0E40-A92A-11d1-B17B-0020AFED142E}" 'LH Margret

	// Now we can make the MSAgent character use the right engine
			BSTR bszTTSModeID = SysAllocString(ppszW);
			hRes = pCharacterEx->SetTTSModeID(bszTTSModeID );
			SysFreeString(bszTTSModeID);
}
