/* Copyright (c) 2003, Gary Dubuque
 * Copyright (c) 2003-2004, Kino Coursey (Mods for CyN)
 *
 * This file is part of AIMLRight.
 * AIMLRight is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation; either version 2, or (at your option) any later 
 * version.
 *
 * AIMLRight is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.

 * You should have received a copy of the GNU General Public License along 
 * with AIMLRight; see the file COPYING. If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

// AIML.cpp: implementation of the AIML Graphmaster classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AIMLpad.h"
#include "nodes.h"
#include <stdlib.h>

extern CAIMLpadApp theApp;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
/////////////////////////////////////////////////////////////////////////////
// TemplateLink

TemplateLink::TemplateLink()
{
	Next=NULL;
	templatepos=0;
}

TemplateLink::~TemplateLink()
{
	//RemoveAll();
}

void TemplateLink::Add(fpos_t new_template)
{

	 if (templatepos==0)
	 {
		 templatepos=new_template;
	 }
	 else
	 {
		if (Next==NULL) Next = new TemplateLink;
		Next->Add(new_template);
	 }
 
}

void TemplateLink::RemoveAll()
{
	TemplateLink *pLink;
	TemplateLink *pOld;

	 pLink = this->Next;
     while (pLink !=NULL) 
	 {
		 pOld = pLink;
		 pLink=pLink->Next;
		 delete pOld;
	 }
}
/////////////////////////////////////////////////////////////////////////////
// Node

Node::Node()
{
	TemplateChain=NULL;
}

Node::~Node()
{
	if (TemplateChain != NULL)
	{
		TemplateChain->RemoveAll();
		delete TemplateChain;
	}
	RemoveAll();
}

void Node::Add(CString key)
{
	Key = key;
	curWord = new Node();

	Node *pTemp = m_WordNodes[key];
	delete pTemp;

	m_WordNodes[key] = curWord;
}

void Node::RemoveAll()
{
	POSITION pos = m_WordNodes.GetStartPosition();
	while (pos != NULL) {
		CString str;
		Node *pNode;
		m_WordNodes.GetNextAssoc(pos,str,pNode);
		delete pNode;
	}
	m_WordNodes.RemoveAll();
}

bool Node::Remove(CString key)
{
	Node *pWord;
	if (!m_WordNodes.Lookup(key,pWord)) return false;
	m_WordNodes.RemoveKey(key);
	delete pWord;
	return true;
}

bool Node::Find(CString key)
{
	if (!m_WordNodes.Lookup(key,curWord)) return false;
	Key = key;
	return true;
}

bool Node::GetFirst()
{
	At = m_WordNodes.GetStartPosition();
	if (At == NULL) return false;
	m_WordNodes.GetNextAssoc(At,Key,curWord);
	return true;
}

bool Node::GetNext()
{
	if (At == NULL) return false;
	m_WordNodes.GetNextAssoc(At,Key,curWord);
	if (At == NULL) return false;
	return true;
}

//
//  Match results class
//

Match::Match()
{
	m_node = NULL;
}

Match::~Match()
{
}

void Match::addInputStar(CString s) {
	inputStar.AddTail(s);
}

void Match::addThatStar(CString s) {
	thatStar.AddTail(s);
}

void Match::addTopicStar(CString s) {
	topicStar.AddTail(s);
}

CString Match::getInputStar(int i) {
	if (i < 1 || i > inputStar.GetCount()) return "";
	POSITION at = inputStar.FindIndex(inputStar.GetCount() - i);
	return inputStar.GetAt(at);
}

CString Match::getThatStar(int i) {
	if (i < 1 || i > thatStar.GetCount()) return "";
	POSITION at = thatStar.FindIndex(thatStar.GetCount() - i);
	return thatStar.GetAt(at);
}

CString Match::getTopicStar(int i) {
	if (i < 1 || i > topicStar.GetCount()) return "";
	POSITION at = topicStar.FindIndex(topicStar.GetCount() - i);
	return topicStar.GetAt(at);
}

int Match::getInputStarCount() {
	return inputStar.GetCount();
}

int Match::getThatStarCount() {
	return thatStar.GetCount();
}

int Match::getTopicStarCount() {
	return topicStar.GetCount();
}


//////////////////////////////////////////////////////////////////////
// Graphmaster: the whole binary tree
//////////////////////////////////////////////////////////////////////

idx::idx()
{
	root = new Node();
	m_size = 0;
}

idx::~idx()
{
	delete root;
}

bool idx::Add(char *str,fpos_t value)
{
	CString path = str;
	Node *leaf;

	leaf = add(root,path);
	leaf->m_template = value; // Original Overwrite

	// new merge
	if (leaf->TemplateChain ==NULL) leaf->TemplateChain=new TemplateLink;
	leaf->TemplateChain->Add(value);

	m_size++;
	return true;
}


Node *idx::add(Node *parent,CString str) {
	CString tail;
	CString word;
	int len = str.GetLength();
	int at = 0;
	int thru;

	for (at = 0;at < len;at++) if (str[at] != ' ') break;

	if (at >= len) return parent;

	for (thru = at;thru < len;thru++) if (str[thru] == ' ') break;

	word = str.Mid(at,thru-at);
	tail = str.Mid(thru);

	if (!parent->Find(word)) parent->Add(word);
	return add(parent->curWord,tail);
}

Match *idx::Select(CString path,Core *World)
{
	// Match *m = select(root,root,1,path,"","");
	World->stars_stkidx=0;
	World->thatstars_stkidx=0;
	World->topicstars_stkidx=0;

	Match *m = select2(root,root,1,path,"","",World);
	return m;
}

Match *idx::select(Node *parent,Node *node,int part,CString str,CString star,CString path)
{
	CString tail;
	CString word;
	int len = str.GetLength();
	int at = 0;
	int thru;
	
	Match *m = NULL;

	for (at = 0;at < len;at++) if (str[at] != ' ') break;

	if (at >= len) {
		if (node->m_template > 0) {
			m = new Match();
			m->m_node = node;
			addStar(m,star,3);
			addPath(m,path,3);
		}
		return m;
	}

	for (thru = at;thru < len;thru++) if (str[thru] == ' ') break;

	word = str.Mid(at,thru - at);
	tail = str.Mid(thru);
	
	if (!word.Compare(_T("<that>")) || !word.Compare(_T("<topic>"))) {
		if (node->Find(word)) {
			m = select(node,node->curWord,part + 1,tail,"","");
			if (m != NULL) {
				addStar(m,star,part);
				addPath(m,path,part);
			}
		}
		return m;
	}

	if (node->Find("_")) {
		m = select(node,node->curWord,part,tail,word,path + " _");
		if (m != NULL) {
			addStar(m,star,part);
			return m;
		}
	}

	if (word.Compare(_T("_")) && word.Compare(_T("*"))) {
		if (node->Find(word)) {
			m = select(node,node->curWord,part,tail,star,path + " " + word);
			if (m != NULL) return m;
		}
	}

	if (node->Find("*")) {
		m = select(node,node->curWord,part,tail,word,path + " *");
		if (m != NULL) {
			addStar(m,star,part);
			return m;
		}
	}

	if (!node->Key.Compare(_T("_")) || !node->Key.Compare(_T("*"))) {
		return select(parent,node,part,tail,star + " " + word,path);
	}

	return NULL;
}

bool idx::testguard(Match *m,Core *World)
{
	fpos_t templpos;
	bool ans;
	char *p1;
	char *p2;
	char *reply;
	TemplateLink *pLink;
	bool found;
	//int i,j;

	ans=true;

	// Create the environment
	if (m == NULL) return TRUE;
	// Have to go through the chain until we find one that either matches or 
	// has no guard. System will then make the Node m_template the one that matches
    pLink = m->m_node->TemplateChain ;
    found=FALSE;


	while ((pLink != NULL) &&( !found))
	{
		// Create local environment for guard evaluation
		/*
		j = Root->getInputStarCount();
		if (j > MAX_STARS) j = MAX_STARS;
		for (i = 0;i < j;i++) sprintf(World->stars[i],"%s",Root->getInputStar(i));
		j = Root->getThatStarCount();
		if (j > MAX_STARS) j = MAX_STARS;
		for (i = 0;i < j;i++) sprintf(World->thatstars[i],"%s",Root->getThatStar(i));
		j = Root->getTopicStarCount();
		if (j > MAX_STARS) j = MAX_STARS;
		for (i = 0;i < j;i++) sprintf(World->topicstars[i],"%s",Root->getTopicStar(i));
		*/

		//templpos = m->m_node->m_template;
			templpos = pLink->templatepos ;
			m->m_node->m_template=templpos;

		// Now peek into the future template to interrogate the <guard>
		if (templpos==0)
		{
			return TRUE;
		}

		char *line = new char[MAX_BUF_LEN];
		char *guard = new char[MAX_BUF_LEN];
		if (templpos) {
			FILE *tmpl = fopen(World->fntmp,"r");
			fsetpos(tmpl,&templpos);

			fgets(line,MAX_BUF_LEN,tmpl);
			rtrim(line);
			expand(line);
			fclose(tmpl);
		} else *line = 0;
		
		 //extractguard(line,guard);
		 p1=strstr(line,"<guard>");
		 p2=strstr(line,"</guard>");
		 if ((p1==NULL)||(p2==NULL)||(p2<p1))
		 {
		   delete guard;
		   delete line;
		   return TRUE;
		 }

		 // real simple extraction from line
		 *p2=0; 
		 strcpy(guard,p1+7); 
		 delete line; // early toss
 
		//ans =evaluateguard(guard);
		reply = World->doCycSystem(guard);
		if (strstr(reply,"FALSE")!=NULL) 
			ans=FALSE;
		else
		{
			ans=TRUE;
			found=TRUE;
		}
		delete guard;
		pLink=pLink->Next;
	}

 return ans;

}

Match *idx::select2(Node *parent,Node *node,int part,CString str,CString star,CString path, Core *World)
{
	// KHC: Adds <guard> taging to the process
	// problem is that the gm system here
	// 1) goes down until it finds a template
	// 2) add the matching patterns it found as it recurses back out
	// The problem with this is we need to evaluate the guard with everything in the matching list
	// but it doesn't generate that until it pops back out !
	// 
	CString tail;
	CString word;
	int len = str.GetLength();
	int at = 0;
	int thru;
	int localstarindex;

	Match *m = NULL;

	for (at = 0;at < len;at++) if (str[at] != ' ') break;

	if (at >= len) {
		if (node->TemplateChain !=NULL )
		{
			m = new Match();
			m->m_node = node;
			addStar(m,star,3);
			addPath(m,path,3);
			localstarindex=pushStar(World,star,3);
			// a template is only valid if the guard matches 
			// otherwise its back to searching
			if (testguard(m,World))
			{
				popStar(World,localstarindex,3);
				return m;
			}
			delete m;
			m=NULL;
		}
		return m;
	}

	for (thru = at;thru < len;thru++) if (str[thru] == ' ') break;

	word = str.Mid(at,thru - at);
	tail = str.Mid(thru);
	
	if (!word.Compare(_T("<that>")) || !word.Compare(_T("<topic>"))) {
		if (node->Find(word)) {
			localstarindex=pushStar(World,star,part);
			m = select2(node,node->curWord,part + 1,tail,"","",World);
			popStar(World,localstarindex,part);
			if (m != NULL) {
				addStar(m,star,part);
				addPath(m,path,part);
			}
		}
		return m;
	}

	if (node->Find("_")) {
		localstarindex=pushStar(World,star,part);
		m = select2(node,node->curWord,part,tail,word,path + " _",World);
		popStar(World,localstarindex,part);
		if (m != NULL) {
			addStar(m,star,part);
			return m;
		}
	}

	if (word.Compare(_T("_")) && word.Compare(_T("*"))) {
		if (node->Find(word)) {
			m = select2(node,node->curWord,part,tail,star,path + " " + word,World);
			if (m != NULL) return m; 
		}
	}

	if (node->Find("*")) {
		localstarindex=pushStar(World,star,part);
		m = select2(node,node->curWord,part,tail,word,path + " *",World);
		popStar(World,localstarindex,part);
		if (m != NULL) {
			addStar(m,star,part);
			return m;
		}
	}

	if (!node->Key.Compare(_T("_")) || !node->Key.Compare(_T("*"))) {
		m = select2(parent,node,part,tail,star + " " + word,path,World);
		return m;
	}

	return NULL;
}


void idx::addStar(Match *m,CString value,int part) {
	value.TrimLeft();
	value.TrimRight();
	if (value.IsEmpty()) return;
	switch (part) {
		case 1: m->addInputStar(value); return;
		case 2: m->addThatStar(value);  return;
		case 3: m->addTopicStar(value);
	}
	return;
}


void idx::addPath(Match *m,CString value,int part) {
	value.TrimLeft();
	value.TrimRight();
	if (value.IsEmpty()) return;
	switch (part) {
		case 1: m->inputPattern = value; return;
		case 2: m->thatPattern  = value; return;
		case 3: m->topicPattern = value;
	}
	return;
}


int idx::pushStar(Core *World,CString value,int part) {
	value.TrimLeft();
	value.TrimRight();
	if (value.IsEmpty()) return 0;
	switch (part) {
		case 1: sprintf(World->stars[World->stars_stkidx++],"%s",value); return( World->stars_stkidx -1);
		case 2: sprintf(World->thatstars[World->thatstars_stkidx++],"%s",value); return(World->thatstars_stkidx-1);
		case 3: sprintf(World->topicstars[World->topicstars_stkidx++],"%s",value); return(World->topicstars_stkidx-1);
	}
	return 0;
}

void idx::popStar(Core *World,int value,int part) {
	switch (part) {
		case 1: World->stars_stkidx =value;return;
		case 2: World->thatstars_stkidx =value;  return;
		case 3: World->topicstars_stkidx =value;
	}
	return;
}



bool idx::Load(char *fname)
{
	CStatusBar* status_bar = (CStatusBar *)theApp.GetMainWnd()->GetDlgItem(AFX_IDW_STATUS_BAR);
	char line[MAX_BUF_LEN];
	bool rtn = false;
	TCHAR szStatus[1024];
	char wrk[20];
	char *pos;
	fpos_t tpos;

	status_bar->SetWindowText(_T("Loading bot memory..."));

	FILE *fpat = fopen(fname,"r");
	if (fpat) {
		while (!feof(fpat)) {
			fgets(line,MAX_BUF_LEN,fpat);
			if (!strncmp(line,"<eof>",5)) break;

			if (!(pos = strstr(line,"@"))) continue;
			strcpy(wrk,pos + 1);
			rtrim(wrk);
			*pos = 0;
			tpos = atol(wrk);

			if (strstr(line,"[that]")) {
				replace(line,"[that]"," <that> ");
				if (strstr(line,"[topic]")) replace(line,"[topic]"," <topic> ");
				else strcat(line," <topic> *");
			} else {
				if (strstr(line,"[topic]")) replace(line,"[topic]"," <that> * <topic> ");
				else strcat(line," <that> * <topic> *");
			}

			while (strstr(line,"<name/>")) replace(line,"<name/>",bot->getbot("name"));
			while (strstr(line,"<bot/>")) replace(line,"<bot/>",bot->getbot("name"));

			if (!Add(line,tpos)) {
				status_bar->SetWindowText(_T("Error in Loading Graphmaster."));
				return false;
			}
		}
		fclose(fpat);

		wsprintf(szStatus,_T("Finished Learning AIML. (%d categories)"),m_size);
		status_bar->SetWindowText(szStatus);

		return true;
	} 
	status_bar->SetWindowText(_T("Patterns file not found."));
	return false;
}
