// rules.h: interface for the Rules class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_RULES_H__23A4ED18_2229_4857_A74A_BD1156C0C3B4__INCLUDED_)
#define AFX_RULES_H__23A4ED18_2229_4857_A74A_BD1156C0C3B4__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Afxtempl.h"
#include "AskDlg.h"

class Subjects;

class Attribute : public CObject
{
public:
	Attribute();
	Attribute(char *name,char *title,char *prompt,char *def,char *parms);

// Attributes
public:
	CString m_name;
	CString m_title;
	CString m_prompt;
	CString m_default;
	CString m_parms;

// Operations
public:
	void FormatAttribute(CString& str);
	void FileAttribute(CString& str);
	void XMLAttribute(CString& str);

// Implementation
public:
	virtual ~Attribute();
};

typedef CTypedPtrMap<CMapStringToOb,CString,Attribute*> CMapStringToAttribute;


class Attributes  
{
public:
	Attributes();

// Operations
public:
	void Add(char *name,char *title,char *prompt,char *def,char *parms);
	void RemoveAll();
	bool Remove(char *name);
	bool Find(char *name);
	bool Seek(CString name);
	bool GetFirst();
	bool GetNext();
	void ListAttributes(CString& str);
	void FormatAttributes(CString& str);
	void FileAttributes(CString& str);
	void XMLAttributes(CString& str);

// Attributes
public:
	CMapStringToAttribute m_mapStringToAttribute;

	POSITION At;
	CString Key;
	CString Name;
	CString Title;
	CString Prompt;
	CString DefaultValue;
	CString Parms;

// Implementation
public:
	virtual ~Attributes();
};

class Premise : public CObject
{
public:
	Premise();
	Premise(char *attr,char *relation,char *value);

// Attributes
public:
	CString m_attribute;
	CString m_relation;
	CString m_value;

// Operations
public:
	void FormatPremise(CString& str);
	void FilePremise(CString& str);
	void XMLPremise(CString& str);

// Implementation
public:
	virtual ~Premise();
};

typedef CTypedPtrArray<CObArray,Premise*> CPremiseArray;

class Result : public CObject
{
public:
	Result();
	Result(char *attr,char *value,double cf);

// Attributes
public:
	CString m_attribute;
	CString m_value;
	double m_cf;

// Operations
public:
	void FormatResult(CString &str);
	void XMLResult(CString &str);
	void FileResult(CString &str);

// Implementation
public:
	virtual ~Result();
};

typedef CTypedPtrArray<CObArray,Result*> CResultArray;

class Rule : public CObject
{
public:
	Rule();
	Rule(char *ruleid,char *subject);

// Attributes
public:
	CString m_ruleid;
	CString m_subject;
	CPremiseArray m_premises;
	CResultArray m_results;

// Operations
public:
	SpecRule(CString& str);
	FormatRule(CString& str);
	FileRule(CString& str);
	XMLRule(CString& str);
	AddPremise(char *attr,char *relation,char *value);
	AddResult(char *attr,char *value,double cf);

// Implementation
public:
	virtual ~Rule();
};

typedef CTypedPtrMap<CMapStringToOb,CString,Rule*> CMapStringToRule;

class Rules  
{
public:
	Rules();

// Operations
public:
	void Add(char *ruleid,char *subject);
	void RemoveAll();
	bool Remove(char *ruleid);
	bool Seek(CString ruleid);
	bool Find(char *ruleid);
	bool GetFirst();
	bool GetNext();
	void ListRules(CString &str);
	void FormatRules(CString &str);
	void XMLRules(CString &str);
	void FileRules(CString &str);

	int SeekTarget(CString s);
	int AssertTarget(CString s);
	double CheckRule(Rule *pRule);
	double CheckPremise(CString a,CString op,CString v,CString r);
	void CompareOp(CString a, CString op, CString v,double cf);
	void PromptAttribute(CString a, CString r, CString &s);
	void Why(CString &str);
	void What(CString r,CString &str);
	void RuleStatus(CString r,CString &str);
	void FormatParameter(CString a,CString r,CString v,CString &s);
	void ExplainPremise(CString a,CString r,CString v,double cf,CString x,int i,CString &s);
	void Which(CString r,CString &str);
	void How(CString r,CString &str);
	void Evaluate(CString r,CString &str);

// Attributes
public:
	CMapStringToRule m_mapStringToRule;

	POSITION At;
	CString Key;
	CString Ruleid;
	CString Subject;
	Rule *curRule;

	Subjects *kb;
	CAskDlg *m_pAskDialog;
	Attributes *atr;

	bool bflg;	// bailout of expert flag

	CStringList goals;
	CStringList atrule;
	int wcnt;	// count of number of whys requested so far for this attribute ask

	double tCF;
	double fCF;

// Implementation
public:
	virtual ~Rules();

};

#endif // !defined(AFX_RULES_H__23A4ED18_2229_4857_A74A_BD1156C0C3B4__INCLUDED_)
