/* Copyright (c) 2000-2002, Jacco Bikker
 * Copyright (c) 2002-2003, Gary Dubuque
 * Copyright (c) 2003-2004, Kino Coursey (Mods for CyN)
 *
 * This file is part of NP+AIML.
 * NP+AIML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation; either version 2, or (at your option) any later 
 * version.
 *
 * NP+AIML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.

 * You should have received a copy of the GNU General Public License along 
 * with NP+AIML; see the file COPYING. If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "support.h"

static char filename[MAX_PATH_LEN];
static unsigned short wfilename[MAX_PATH_LEN];

static bool nocompress;

// operating system reoutines

int getclock()
{
	return GetTickCount();
}

char *getdts(char *buf)
{
	SYSTEMTIME temp;
	temp.wHour = 0;
	temp.wMinute = 0;
	temp.wSecond = 0;
	GetLocalTime(&temp);
	sprintf(buf,"%d/%d/%d %d:%02d:%02d",temp.wMonth,temp.wDay,temp.wYear,temp.wHour,temp.wMinute,temp.wSecond);
	return buf;
}

char *getdate(char *buf)
{
	SYSTEMTIME temp;
	temp.wHour = 0;
	temp.wMinute = 0;
	temp.wSecond = 0;
	GetLocalTime(&temp);
	sprintf(buf,"%d/%d/%d",temp.wMonth,temp.wDay,temp.wYear);
	return buf;
}

char *gettime(char *buf)
{
	SYSTEMTIME temp;
	temp.wHour = 0;
	temp.wMinute = 0;
	temp.wSecond = 0;
	GetLocalTime(&temp);
	sprintf(buf,"%d:%02d:%02d",temp.wHour,temp.wMinute,temp.wSecond);
	return buf;
}

void timestamp(CString str)
{
	SYSTEMTIME temp;
	temp.wHour = 0;
	temp.wMinute = 0;
	temp.wSecond = 0;
	GetLocalTime(&temp);
	str.Format(_T("%d/%d/%d %d:%02d:%02d"),temp.wMonth,temp.wDay,temp.wYear,temp.wHour,temp.wMinute,temp.wSecond);
}

void randomize()
{
	SYSTEMTIME ltime;
	GetLocalTime(&ltime);
	srand(getclock());
}

char *fullpath(char *a_File)
{
	GetModuleFileName(0,filename,MAX_PATH_LEN);
	int pos = strlen(filename);
	while (--pos) if (filename[pos] == '\\') break;
	strcpy(filename + pos + 1,a_File);
	return filename;
}

LPWSTR fullpath(LPWSTR a_File)
{
 	USES_CONVERSION;
	GetModuleFileName(0,filename,MAX_PATH_LEN);
	wcscpy(wfilename,A2W(filename));
	int pos = wcslen(wfilename);
	if (pos) while (--pos) if (wfilename[pos] == '\\') break;
	wcscpy(wfilename + pos + 1,a_File);
	return (LPWSTR)wfilename;
}

bool fexists(char* file)
{
	FILE *test = fopen(file,"r");
	if (!test) return false;
	fclose(test);
	return true;
}

int random(int max)
{
	return (int)((float)rand() * max / RAND_MAX);
}

// XML parsing functions

bool XMLtag(char *txt,char *tag)
{
	char *x;
	char *y;

	if (!(x = strstr(txt,"<"))) return false;
	if (!strstr(x,">")) return false;
    for (y = tag;(*y++ = *x) != '>';x++);
	*y = 0;
	return true;
}

bool XMLvalue(char *tag,char *attr,char *value)
{
	char buf[MAX_TAG_LEN];
	char *x;
	char *y;

	strcpy(buf,attr);
	strcat(buf,"=");
	if (!(y = strstr(tag,buf))) return false;
	x = y + strlen(buf);

	if (*x == '\"') {
		for (y = value,x++;*x && *x != '\"';x++) *y++ = *x;
	} else if (*x == '\'') {
		for (y = value,x++;*x && *x != '\'';x++) *y++ = *x;
	} else {
		for (y = value;*x && !strchr(" />",*x);x++) *y++ = *x;
	}
	*y = 0;
	return (*x != 0);
}

bool XMLnode(char *text,char *otag,char *data,char *ctag)
{
	char *x;
	char *y;
	char *z;
	char *at;
	bool fnd;
	bool inquote=false;
	int len;

	if (*text != '<') return false;
	if (!strstr(text,">")) return false;
    for (x = text,y = otag;(*y++ = *x) != '>';x++); *y = 0;

	if (*(y - 2) == '/') {
		*data = 0;
		*ctag = 0;
		return true;
	}
	at = x;
	while (z = strstr(at,"<")) {
		if (!strstr(z,">")) return false;
		if (*(z + 1) == '/') break;
		char *otag1 = new char[MAX_TAG_LEN];
		char *data1 = new char[MAX_SIZE];
		char *ctag1 = new char[MAX_TAG_LEN];
		fnd = XMLnode(z,otag1,data1,ctag1);
		if (fnd) len = strlen(otag1) + strlen(data1) + strlen(ctag1);
		delete ctag1;
		delete data1;
		delete otag1;
		if (!fnd) return false;
		at = z + len - 1;
	}
	if (!z) return false;
	for (x++,y = data;x < z;x++) *y++ = *x; *y = 0;
	for (y = ctag;(*y++ = *z) != '>';z++);  *y = 0;
	return true;
}

bool XMLnode2(char *text,char *otag,char *data,char *ctag)
{
	char *x;
	char *y;
	char *z;
	char *at;
	bool fnd;
	bool inquote=false;
	int len;

	if (*text != '<') return false;
	if (!strstr(text,">")) return false;
    for (x = text,y = otag;(((*y++ = *x) != '>')||(inquote));x++)
	{
		if (*x=='\"')
		{
			inquote=!inquote;
		}
	}
	*y = 0;

	if (*(y - 2) == '/') {
		*data = 0;
		*ctag = 0;
		return true;
	}
	at = x;
	while (z = strstr(at,"<")) {
		if (!strstr(z,">")) return false;
		if (*(z + 1) == '/')
		{
			if( *(z+2)==*(text+1)) break;
		}
		char *otag1 = new char[MAX_VALUE_LEN];
		char *data1 = new char[MAX_SIZE];
		char *ctag1 = new char[MAX_VALUE_LEN];
		fnd = XMLnode2(z,otag1,data1,ctag1);
		if (fnd) len = strlen(otag1) + strlen(data1) + strlen(ctag1);
		delete ctag1;
		delete data1;
		delete otag1;
		if (!fnd) return false;
		at = z + len - 1;
	}
	if (!z) return false;
	for (x++,y = data;x < z;x++) *y++ = *x; *y = 0;
	for (y = ctag;(*y++ = *z) != '>';z++);  *y = 0;
	return true;
}

// text manipulation

void remove(char *text,char *first,char *last)
{
	char *pos1, *pos2;
	if (!(pos1 = strstr(text,first))) return;
	if (!(pos2 = strstr(pos1,last))) return;
	*pos1 = 0;
	strcat(pos1,pos2 + strlen(last));
}

bool replace(char *line,char *string1,char *string2)
{
	char buf[MAX_SIZE];
	char *pos = strstr(line,string1);
	if (pos) {
		strcpy(buf,pos + strlen(string1));
		strcpy(pos,string2);
		strcat(pos,buf);
		return true;
	}
	return false;
}

void split(char *line,char *text,int cnt)
{
	char *at;
	char *pos = line;
	int i;

	if (cnt == 0) { strcpy(line,text); return; }

	for (at = text,i = 0;*at;at++) {
		if (*at == '.' || *at == ';' || *at == '?' || *at == '!') i++;
	}
	for (at--;at > text;at--) if (*at > ' ') break;
	if (*at != '.' && *at != ';' && *at != '?' && *at != '!') i++;
	for (at = text;*at && i >= cnt;at++) {
		if (*at == '.' || *at == ';' || *at == '?' || *at == '!') {
			i--;
			while (*at && *at <= ' ') at++;
			continue;
		}
		if (i == cnt) { if (*at >= ' ') *pos++ = *at; }
	}
	*pos = 0;
}

void cleanspace(char *text)
{
	char a = ' ';
	char *b = text;

	for (char *c = text;*c;c++) {
		if (*c == ' ') {
			if (a != *c) *b++ = *c;
			a = *c;
		} else if (*c < ' ') { a = ' ';
		} else a = *b++ = *c;
	}
	*b = 0;
}

void cleaner(char *text)
{
	char *b = text;

	for (char *c = text;*c;c++) {
		if (isalnum(*c) || strchr(" @-:\'",*c) || (isalnum(*(c + 1)) && *c == '.') ) *b++ = *c;
	}
	*b-- = 0;
	while (b >= text && *b <= ' ') *b-- = 0;
}

void rtrim(char *text)
{
	for (char *c = text + strlen(text);c >= text;c--) if (*c < 32) *c = 0; else break;
}

void fixnewlines(char *line)
{
	while (replace(line,"<br>","\r\n"));
	while (replace(line,"<br/>","\r\n"));
	while (replace(line,"<br />","\r\n"));
	while (replace(line,"<BR>","\r\n"));
	while (replace(line,"<BR/>","\r\n"));
	while (replace(line,"<BR />","\r\n"));
	while (replace(line,"<Br>","\r\n"));
	while (replace(line,"<Br/>","\r\n"));
	while (replace(line,"<Br />","\r\n"));
}

void uppercase(char *text)
{
	for (char *c = text;*c;c++)	if (islower(*c)) *c += 'A' - 'a';
}

void lowercase(char *text)
{
	for (char *c = text;*c;c++)	if (isupper(*c)) *c -= 'A' - 'a';
}

void propercase(char *text)
{
	bool flg = true;

	for (char *c = text;*c;c++)	{
		if (islower(*c)) {
			if (flg) *c += 'A' - 'a';
			flg = false;
		} else if (isupper(*c)) {
			if (flg) flg = false; else *c -= 'A' - 'a';
		} else flg = true;
	}
}

// text comparisons

int strpcmp(char *s1,char *s2)
{
	char *a = s1;
	char *b = s2;
	char c1, c1x;
	char c2, c2x;

	if (!*a && !*b) return 0;

	c1 = *a;
	c2 = *b;
	if (islower(c1)) c1 += 'A' - 'a';
	if (islower(c2)) c2 += 'A' - 'a';
	if (c1 != c2) {
		if (c1 == '_') return -1;
		if (c2 == '_') return 1;
		if (c1 == '*') return 1;
		if (c2 == '*') return -1;
		if (isalpha(c1)) {	if (isalpha(c2) && (c1 > c2)) return 1; }
		else if (isalpha(c2)) return 1;
		else if (c1 == '[') return -1;
		else if (c2 == '[') return 1;
		else if (c1 > c2) return 1;
		return -1;
	}
	c1x = c1;
	c2x = c2;
	a++;
	b++;

	while (*a || *b) {
		c1 = *a;
		c2 = *b;
		if (islower(c1)) c1 += 'A' - 'a';
		if (islower(c2)) c2 += 'A' - 'a';
		if (c1 != c2) {
			if (c1 == '_') return -1;
			if (c2 == '_') return 1;
			if (c1 == '[') return -1;
			if (c2 == '[') return 1;
			if (c1 == '*') return 1;
			if (c2 == '*') return -1;
			if (c1x == '*') if (c1 == '@' || c1 == '!' || c1 == '[') return 1;
			if (c2x == '*') if (c2 == '@' || c2 == '!' || c2 == '[') return -1;
			if (c1x == '_') if (c1 == '@' || c1 == '!' || c1 == '[') return 1;
			if (c2x == '_') if (c2 == '@' || c2 == '!' || c2 == '[') return -1;
			if (c1 > c2)   return 1;
			return -1;
		}
		a++;
		b++;
		c1x = c1;
		c2x = c2;
	}
	return 0;
}

int stricmp(char *s1,char *s2)
{
	char *a = s1;
	char *b = s2;
	char c1;
	char c2;

	while (*a || *b) {
		c1 = *a;
		c2 = *b;
		if (isupper(c1)) c1 -= 'A' - 'a';
		if (isupper(c2)) c2 -= 'A' - 'a';
		if (c1 != c2) { if (c1 < c2) return -1; else return 1; }
		a++;
		b++;
	}
	return 0;
}

int strnicmp(char *s1,char *s2,int cnt)
{
	char *a = s1;
	char *b = s2;
	char c1;
	char c2;

	for (int i = 0;*a || *b;i++) {
		if (i == cnt) break;
		c1 = *a;
		c2 = *b;
		if (isupper(c1)) c1 -= 'A' - 'a';
		if (isupper(c2)) c2 -= 'A' - 'a';
		if (c1 != c2) { if (c1 < c2) return -1; else return 1; }
		a++;
		b++;
	}
	return 0;
}

char *stristr(char *s1,char *s2)
{
	char *a;
	char *b;
	char *c;
	char c1;
	char c2;

	for (a = s1;*a;a++) {
		for (b = s2,c = a;*b && *c;b++,c++) {
			c1 = *b;
			c2 = *c;
			if (isupper(c1)) c1 -= 'A' - 'a';
			if (isupper(c2)) c2 -= 'A' - 'a';
			if (c1 != c2) break;
		}
		if (!*b) return a;
	}
	return 0;
}

bool patrns(char *line,char *best)
{
	if (*best == '_' && *line != '_') return true;
	if (*line == '*' && *best != '*') return true;
	if (*line == *best) {
		if (strstr(line,"*") && !strstr(best,"*") && !strstr(best,"_")) return true;
		if (strlen(line) < strlen(best)) { if (!strnicmp(line,best,strlen(line))) return true; }
		if (stricmp(line,best) < 0) return true;
	}
	return false;
}

// Space saving optimizations

void expand(char *line)
{
	if (nocompress) return;

	while (replace(line,"~}","</srai>"));
	while (replace(line,"~{","<srai>"));
	while (replace(line,"~%","<random><li>"));
	while (replace(line,"~&","</li><li>"));
	while (replace(line,"~@","</li></random>"));
	while (replace(line,"~Q","</set>"));
	while (replace(line,"~M"," can "));
	while (replace(line,"~L","index="));
	while (replace(line,"~K","name="));
	while (replace(line,"~J","ell me "));
	while (replace(line,"~I","<usesaved/>"));
	while (replace(line,"~H"," favorite "));
	while (replace(line,"~G"," about "));
	while (replace(line,"~F","WHAT "));
	while (replace(line,"~E","I WAS "));
	while (replace(line,"~D","IT IS "));
	while (replace(line,"~C","I HAVE "));
	while (replace(line,"~B","I DID "));
	while (replace(line,"~A","I AM "));
	while (replace(line,"~z","<ironic/>"));
	while (replace(line,"~y","<upset/>"));
	while (replace(line,"~x","<worried/>"));
	while (replace(line,"~w","<serious/>"));
	while (replace(line,"~U","</save>"));
	while (replace(line,"~u","<save>"));
	while (replace(line,"~t","<suspecting/>"));
	while (replace(line,"~s","<angry/>"));
	while (replace(line,"~r","<getname/>"));
	while (replace(line,"~q","<set "));
	while (replace(line,"~p","<get "));
	while (replace(line,"~o","</think>"));
	while (replace(line,"~n","<think>"));
	while (replace(line,"~m","<star/>"));
	while (replace(line,"~l","<thinking/>"));
	while (replace(line,"~k","<interested/>"));
	while (replace(line,"~j","<happy/>"));
	while (replace(line,"~i","<glad/>"));
	while (replace(line,"~h","<funny/>"));
	while (replace(line,"~g","<puzzled/>"));
	while (replace(line,"~f","I don't "));
	while (replace(line,"~e","I am "));
	while (replace(line,"~d","o you "));
	while (replace(line,"~c"," you"));
	while (replace(line,"~b","<sr/>"));
	while (replace(line,"~a","<person/>"));
	while (replace(line,"~)","</system>"));
	while (replace(line,"~(","<system>"));
}

void compress(char *line)
{
	if (nocompress) return;

	while (replace(line,"<srai>","~{"));
	while (replace(line,"</srai>","~}"));
	while (replace(line,"<random><li>","~%"));
	while (replace(line,"</li><li>","~&"));
	while (replace(line,"</li></random>","~@"));
	while (replace(line,"<system>","~("));
	while (replace(line,"</system>","~)"));
	while (replace(line,"<person/>","~a"));
	while (replace(line,"<sr/>","~b"));
	while (replace(line," you","~c"));
	while (replace(line,"o you ","~d"));
	while (replace(line,"I am ","~e"));
	while (replace(line,"I don't ","~f"));
	while (replace(line,"<puzzled/>","~g"));
	while (replace(line,"<funny/>","~h"));
	while (replace(line,"<glad/>","~i"));
	while (replace(line,"<happy/>","~j"));
	while (replace(line,"<interested/>","~k"));
	while (replace(line,"<thinking/>","~l"));
	while (replace(line,"<star/>","~m"));
	while (replace(line,"<think>","~n"));
	while (replace(line,"</think>","~o"));
	while (replace(line,"<get ","~p"));
	while (replace(line,"<set ","~q"));
	while (replace(line,"</set>","~Q"));
	while (replace(line,"<getname/>","~r"));
	while (replace(line,"<angry/>","~s"));
	while (replace(line,"<suspecting/>","~t"));
	while (replace(line,"<save>","~u"));
	while (replace(line,"</save>","~U"));
	while (replace(line,"<serious/>","~w"));
	while (replace(line,"<worried/>","~x"));
	while (replace(line,"<upset/>","~y"));
	while (replace(line,"<ironic/>","~z"));
	while (replace(line,"I AM ","~A"));
	while (replace(line,"I DID ","~B"));
	while (replace(line,"I HAVE ","~C"));
	while (replace(line,"IT IS ","~D"));
	while (replace(line,"I WAS ","~E"));
	while (replace(line,"WHAT ","~F"));
	while (replace(line," about ","~G"));
	while (replace(line," favorite ","~H"));
	while (replace(line,"<usesaved/>","~I"));
	while (replace(line,"ell me ","~J"));
	while (replace(line,"name=","~K"));
	while (replace(line,"index=","~L"));
	while (replace(line," can ","~M"));
}

void setcompress(bool val)
{
	nocompress = !val;
}

void newstr(char **s1,char *s2)
{
	*s1 = new char[strlen(s2) + 2];
	strcpy(*s1,s2);
}

bool isNumeric(CString s)
{
	int i;
	bool bPeriod = false;
	bool bSign = false;
	char c;

	for (i = 0;i < s.GetLength();i++) {
		c = (char) s[i];
		if (c >= '0' && c <= '9') { bSign = true; continue; }
		if (c == '.') {
			if (bPeriod) return false;
			bPeriod = true;
			continue;
		}
		if (c == '-') {
			if (bSign) return false;
			bSign = true;
			continue;
		}
		if (c == '+') {
			if (bSign) return false;
			bSign = true;
			continue;
		}
		return false;
	}
	return true;
}


//////////////////////////////////////////////////////////////
// CLIPWARE SERVER/CLIENTS
// Simplest possible server/clients to just make the connection
//
// Better solutions greatly appereciated !!!!
//
// And thanks to the origial authors 
//     (forgive us for what we were forced to do...)
//
//////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////
// INITIAL IRC CLIENT ATTEMPT
// Bolted in from Aine bot project \source\extras_old\irc_old
//
///////////////////////////////////////// ///////////////////
#ifdef WIN32
	#include <winsock.h>
	#include <winbase.h>
#define EINPROGRESS WSAEINPROGRESS
#define EWOULDBLOCK WSAEWOULDBLOCK
#define EISCONN     WSAEISCONN

#else
	#include <unistd.h>
	#include <netdb.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#ifdef SUNOS
		#include <sys/varargs.h>
	#else
		#include <stdarg.h>
	#endif
#endif

//==============================================================
// ADAPTED FROM :
// Charlotte - A small Windows web server 
//
// By Stuart Patterson
//==============================================================
#include <windows.h>
#include <process.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>


//--------------------------------------------------------------
// manifest constants
//--------------------------------------------------------------


//--------------------------------------------------------------
// structures
//--------------------------------------------------------------

//--------------------------------------------------------------
// prototypes for WWW
//--------------------------------------------------------------
SOCKET StartWebServer();
//int WaitForClientConnections(SOCKET server_socket);
void HandleHTTPRequest(char *inbuf,	HTTPRequestHeader *requestheaderP, void *data );
int findMimeType(const char *extension);
BOOL ParseHTTPHeader(char *receivebuffer, HTTPRequestHeader *requestheaderP);
void OutputHTTPError(SOCKET client_socket, int statuscode );
void OutputHTTPRedirect(SOCKET client_socket, const char *defaulturl);
void OutputScreenError(const char *errmsg);
int SocketRead(SOCKET client_socket, char *receivebuffer, int buffersize);
void DetermineHost( char *hostname );


//--------------------------------------------------------------
// global vars
//--------------------------------------------------------------
int wwwport = 81;
int wwwactive=0;
char wwwroot[MAX_PATH];
char wwwpage[MAX_PATH];
char hostname[MAX_PATH];

SOCKET www_server_socket;

CRITICAL_SECTION output_criticalsection;


MimeAssociation mimetypes[] = { 
	{ ".txt", "text/plain" },
	{ ".html","text/html" },
	{ ".htm", "text/html" },
	{ ".gif", "image/gif" },
	{ ".jpg", "image/jpeg" }
};
//------------------------------------------------------------
int serv = -1;
in_parsed irc_in;

//You must change the below to reflect your defaults.  

char ircnick[128] = "Cynix";
char ircserver[255] = "irc.freenode.net";
char ircroom[32] = "#logicmoo";   //note, you can put multiple rooms here.
int ircport = 6667; 
char* var_file = new char[64];
char* scrap = new char[512];
#define MIN_IRC_WAIT 4
WSADATA wsaData;
int winsock_active =0;


//-------------------------------------------------------------




int writeit(int server, char *fmt, ...)
{
  va_list va;
  char *buf = new char[MAX_BUF_LEN];
  va_start(va,fmt);
  vsprintf(buf,fmt,va);
  va_end(va);
#ifdef WIN32
  if (send(server,buf,strlen(buf),0) == -1)
#else  
  if (write(server,buf,strlen(buf)) == -1)
#endif
  {
    delete(buf);
    return(-1);
  }
  delete(buf);
  return(1);
}


char* spacefixer(char *s,char b[1]) 
{
  unsigned int i,j;
  for (i = 0, j = strcspn(s,b)+1; j <= strlen(s); i++, j++) s[i] = s[j];
  s[i+1]=0;
  return(s);
}

int set_blocking(SOCKET sfd, int block)  /* Not tested. */
{
  u_long noblock = (block ? 0L : 1L);
  int rc;

  rc= ioctlsocket(sfd, FIONBIO, &noblock);

  if(rc) {errno = WSAGetLastError();}

  return(rc);
}
	

void init_winsocks()
{
		int rVal=0;
		
	if (winsock_active!=0) return;
	WORD version = MAKEWORD(1,1);

	// We're compatible with version 1.1 of WS2_32.dll
	WSAStartup(version,(LPWSADATA)&wsaData);
	winsock_active=1;
}
//int call_socket (char hostname[254], unsigned short portnum)
int call_socket (char *hostname, unsigned short portnum)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  int s;
  int iRet;


	init_winsocks();

  if ((hp = gethostbyname (hostname)) == NULL) {    return (-1);
  }
  memset(&sa, 0, sizeof (sa));
  memcpy((char *) &sa.sin_addr, hp->h_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons ((u_short) portnum);

  if ((s = socket (hp->h_addrtype, SOCK_STREAM, 0)) < 0)
    return (-1);
  if (connect (s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
	#ifdef WIN32
		closesocket (s);
	#else
		close (s);
	#endif
    return (-1);
  }
  iRet=set_blocking(s,FALSE);
  return (s);
}

//ok, the plan is to use a linked list to store every word into an array
//of words, then make all these sender, reciever scrap things, point to
//that array. 

in_parsed parseold(char *s)
{
  in_parsed a;
  memset(a.sender,0,124);
  memset(a.reciever,0,24);
  memset(a.method,0,24);
  memset(a.answer,0,24);
  memset(a.sendnick,0,24);
  memset(a.senduser,0,24);
  memset(a.sendip,0,254);
  memset(a.text,0,512);
  char* scrap = new char[512];
  memset(scrap,0,512);
  s = spacefixer(s,":");
  strncpy(a.sender,s,strcspn(s," "));
  scrap = a.sender;
  s = spacefixer(s," ");
  strncpy(a.method,s,strcspn(s," "));
  s = spacefixer(s," ");
  strncpy(a.reciever,s,strcspn(s," "));
  s = spacefixer(s," ");
  if (s[0] != ':'){
    strncpy(a.answer,s,strcspn(s,":"));
  }
  s = spacefixer(s,":");
  strncpy(a.text,s,strcspn(s,"\n"));
  if(strchr(scrap,'!')){
    strncpy(a.sendnick,scrap,strcspn(scrap,"!"));
    scrap = spacefixer(scrap,"!");
    strncpy(a.senduser,scrap,strcspn(scrap,"@"));
    scrap = spacefixer(scrap,"@");
    strncpy(a.sendip,scrap,strcspn(scrap,"\n"));
  }
  return(a);
}


int ircinit()
{
  //initialize();
  //connect to ircserver and cross fingers.
  if (!(serv = call_socket(ircserver,ircport))){
    printf("Connection Failed, Exiting.\n");
    return(0);
  }
  writeit(serv,"PASS null\n");

#ifdef WIN32
  Sleep( (clock_t)MIN_IRC_WAIT * CLOCKS_PER_SEC );
#else
  sleep(MIN_IRC_WAIT);
#endif

  writeit(serv,"NICK %s\n",ircnick);

#ifdef WIN32
  Sleep( (clock_t)MIN_IRC_WAIT * CLOCKS_PER_SEC );
#else
  sleep(MIN_IRC_WAIT);
#endif

  writeit(serv,"USER blah blah@nowhere.com blah.com blah\n");
  memset(scrap,0,512);

#ifdef WIN32
	recv (serv, scrap, 512, 0);
#else
	read(serv,scrap,512);
#endif

  if (strncmp(scrap,"PING",4)==0) {
    scrap=spacefixer(scrap,":");
    writeit(serv,"PONG %s\n",scrap);
  }

#ifdef WIN32
  Sleep( (clock_t)MIN_IRC_WAIT * CLOCKS_PER_SEC );
#else
  sleep(MIN_IRC_WAIT);
#endif

  writeit(serv,"JOIN %s\n",ircroom);
  return(0);
}

int irctick(char *inputbuf)
{
	//int iRet;

  //while (42) {
  //  memset(answer,0,512);
    memset(scrap,0,512);
	#ifdef WIN32
		if(recv(serv,scrap,512,0) < 1){
	#else
    if(read(serv,scrap,512) < 1){      
	#endif
      printf("Connection lost, exiting.\n");
      return(0);
    }
    if (strncmp(scrap,"PING",4)==0) {
      scrap=spacefixer(scrap,":");
      writeit(serv,"PONG %s\n",scrap);
    }
    irc_in = parseold(scrap);
    if (strncmp(irc_in.method,"JOIN",4)==0){
      writeit(serv,"NOTICE %s :Hi! Right now CyN is best using PM. Also if you could send one line at a time please. And none of those crazy actions.  Thanks.\n",irc_in.sendnick);
    }

	#ifdef WIN32
    if ((strncmp(irc_in.text,"\1",1)==0) && (_stricmp(irc_in.reciever,ircnick)==0)){
	#else
    if ((strncmp(irc_in.text,"\1",1)==0) && (strcasecmp(irc_in.reciever,ircnick)==0)){
	#endif

      writeit(serv,"NOTICE %s :I can't recieve that type of message. Please type only plain PMs.  Also Please send one line at a time.\n",irc_in.sendnick);
    }
    else {
		#ifdef WIN32
      if ( !_stricmp(irc_in.reciever,ircnick) && !_stricmp(irc_in.method,"PRIVMSG") && strncmp(irc_in.sendnick,"\0",1)){
		#else
      if ( !strcasecmp(irc_in.reciever,ircnick) && !strcasecmp(irc_in.method,"PRIVMSG") && strncmp(irc_in.sendnick,"\0",1)){
		#endif
        //fflush(stdout);
        //sprintf(var_file,"%s.log",in.sendnick);
        //******* INSERT YOUR BOT CODE HERE *******
		strcpy(inputbuf,irc_in.text);
		return(1);
		//loadvars(var_file);
		//printf("%s:%s\n",in.sendnick,in.text);
		//  answer = respond(in.text);
	//	iRet = bot->getAnswer(in.text ,answer);
        //stripurl(answer);
        //printf("hippie:%s\n",answer);
        //writeit(serv,"PRIVMSG %s :%s\n",in.sendnick,answer);
        //savevars(var_file);
      }
    }
//  wipeparse();
//  }
	  return (0);
}

int ircreply(char *outbuf)
{
	long msglen;
	long i;
	char *p;
	msglen=strlen(outbuf);

    if (strlen(outbuf)<512)
	{
	writeit(serv,"PRIVMSG %s :%s\n",irc_in.sendnick,outbuf);
		return(0);
	}
	else
	{
		char *msgbuf=new char [512];
		for(i=0;i<msglen;i=i+256)
		{
			memset(msgbuf,0,512);
			p=outbuf+i;
			strncpy(msgbuf,p,256);
			writeit(serv,"PRIVMSG %s :%s\n",irc_in.sendnick,msgbuf);
			Sleep( (clock_t)2 * CLOCKS_PER_SEC );
		}
		delete msgbuf;
		return(0);
		
	}
}






//--------------------------------------------------------------
//  www_init()
//		Start of application.  Initializes winsock and starts
//		server
//--------------------------------------------------------------

void decodeURLEncode(char *s) {
  int count;
  int len;
  int i;
  count=0;
  len= strlen(s);
  for (i=0; i<len; i++) {
	  if ((s[i]=='%')&&(s[i+1]=='2')&&(s[i+2]>='0'))
	  {s[count++]=0x20+(s[i+2]-'0');
	   i=i+2;
	  }
	else
	{
		s[count++]=s[i];
	}
  }
  s[count++] = 0;
}

int wwwinit()
{
	int iRet;

	// init the winsock libraries
	init_winsocks();

	// get name of webserver machine. needed for redirects
   // gethostname does not return a fully qualified host name
	DetermineHost(hostname);

	// init the webserver
	www_server_socket = StartWebServer();
	if ( www_server_socket )
	{
		iRet=set_blocking(www_server_socket,FALSE);

		if ( listen(www_server_socket,SOMAXCONN) == SOCKET_ERROR )
		{
			OutputScreenError("Error in listen()");
			closesocket(www_server_socket);
			return(0);
		}
	}

	return(0);
}

//--------------------------------------------------------------
// wwwtick()
//	looks for a pending request. If there is one then process it
//--------------------------------------------------------------
int wwwtick(char *inbuf,ClientInfo *ci,HTTPRequestHeader *requestheaderP)
{
	
	SOCKET server_socket;
	SOCKET client_socket;
	SOCKADDR_IN client_address;
	int client_address_len;
	//ClientInfo *ci;
	int iCt,iNotUsed;
	struct timeval timeout;
	fd_set setRead;

	server_socket=www_server_socket;
	FD_ZERO(&setRead);
	FD_SET(www_server_socket,&setRead);
	timeout.tv_sec =0L;
	timeout.tv_usec =0L;
	iNotUsed=0;
 	client_address_len = sizeof(SOCKADDR_IN);


   // // while we are here loop while there are extern connections to service
   // //while ( (iCt=select(iNotUsed,&setRead,NULL,NULL,&timeout)) !=0)
   if  ( (iCt=select(iNotUsed,&setRead,NULL,NULL,&timeout)) ==0)
	   return(0);
   else
	{
		client_socket = accept(server_socket,(struct sockaddr *)&client_address,&client_address_len);
		if ( client_socket == INVALID_SOCKET )
		{
			OutputScreenError("Error in accept()");
			closesocket(server_socket);
			return(0);
		}

      // copy client ip and socket so the HandleHTTPRequest thread
      // and process the request.
		//ci = new ClientInfo;
		ci->client_socket = client_socket;
		memcpy(&(ci->client_ip),&client_address.sin_addr.s_addr,4);

		// for each request start a new thread!
		//_beginthread(HandleHTTPRequest,0,(void *)ci);
		HandleHTTPRequest(inbuf, requestheaderP,(void *) ci);
		return(1);
	}
	return(0);
}

int wwwreply(char *outbuf, ClientInfo *ci,	HTTPRequestHeader *requestheaderP)
{

	SOCKET client_socket;
	//int size;
	//char receivebuffer[COMM_BUFFER_SIZE];
	char sendbuffer[COMM_BUFFER_SIZE];

	FILE *in;
	char *filebuffer;
	long filesize;
	DWORD fileattrib;
	
	client_socket = ci->client_socket;

	fileattrib = GetFileAttributes(requestheaderP->filepathname);

	if ( fileattrib != -1 && fileattrib & FILE_ATTRIBUTE_DIRECTORY )
	{
		OutputHTTPRedirect(client_socket, requestheaderP->url);
		return(0);
	}

	in = fopen(requestheaderP->filepathname,"rb");  // read binary
	if ( !in )
	{
		// file error, not found?
		OutputHTTPError(client_socket, 404);   // 404 - not found
		return(0);
	}

	// determine file size
	fseek(in,0,SEEK_END);
	filesize = ftell(in);
	fseek(in,0,SEEK_SET);

	// allocate buffer and read in file contents
	filebuffer = new char[filesize+strlen(outbuf)*4];
	fread(filebuffer,sizeof(char),filesize,in);
	fclose(in);
	
	if (requestheaderP->mime_index ==0) 
	{
		if (strstr(filebuffer,"<BOTREPLY/>")!=NULL ) 
		{
			filebuffer[filesize+1]=0; // make sure there is a null
			replace(filebuffer,"<BOTREPLY/>",outbuf); // insert the botreply into the HTML
			filesize=filesize-11 +strlen(outbuf);
		}
	}
	// send the http header and the file contents to the browser
	strcpy(sendbuffer,"HTTP/1.0 200 OK\r\n");
	strncat(sendbuffer,"Content-Type: ",COMM_BUFFER_SIZE);
	strncat(sendbuffer,mimetypes[findMimeType(requestheaderP->filepathname)].mime,COMM_BUFFER_SIZE);
	sprintf(sendbuffer+strlen(sendbuffer),"\r\nContent-Length: %ld\r\n",filesize);
	strncat(sendbuffer,"\r\n",COMM_BUFFER_SIZE);

	send(client_socket,sendbuffer,strlen(sendbuffer),0);
	send(client_socket,filebuffer,filesize,0);


	closesocket(client_socket);
	filebuffer[0]=0;
	delete [] filebuffer;

 return(1);
}
//--------------------------------------------------------------
//	HandleHTTPRequest()
//		Executed in its own thread to handling http transaction
//--------------------------------------------------------------
void HandleHTTPRequest(char *inbuf,	HTTPRequestHeader *requestheaderP, void *data )
{
	SOCKET client_socket;
	//HTTPRequestHeader requestheader;
	int size;
	char receivebuffer[COMM_BUFFER_SIZE];
//	char sendbuffer[COMM_BUFFER_SIZE];

	client_socket = ((ClientInfo *)data)->client_socket;
	requestheaderP->client_ip = ((ClientInfo *)data)->client_ip;
	
	//delete data;

	size = SocketRead(client_socket,receivebuffer,COMM_BUFFER_SIZE);
	if ( size == SOCKET_ERROR || size == 0 )
	{
		OutputScreenError("Error calling recv()");
		closesocket(client_socket);
		return;
	}
	receivebuffer[size] = NULL;

	if ( !ParseHTTPHeader(receivebuffer,requestheaderP) )
	{
		// handle bad header!
		OutputHTTPError(client_socket, 400);   // 400 - bad request
		return;
	}

	if ( strstr(requestheaderP->method,"GET") )
	{
		if (strlen(requestheaderP->url)>1 ) 
		{
		    char *p;
			p = strstr(requestheaderP->url,"?input=");
			if (p!=NULL) 
			{
				strcpy(inbuf,p+7);
				while (strstr(inbuf,"+")!=NULL) replace(inbuf,"+"," ");
			}
			else
				strcpy(inbuf,&requestheaderP->url[1]);
			
		}
		else
				strcpy(inbuf,requestheaderP->url);
		requestheaderP->mime_index = findMimeType(requestheaderP->url);
		if (requestheaderP->mime_index==0)
		{
			// the default page 
			strncpy(requestheaderP->filepathname,wwwroot,MAX_PATH);
			strncat(requestheaderP->filepathname,wwwpage,MAX_PATH);
			_fullpath(requestheaderP->filepathname,requestheaderP->filepathname,MAX_PATH);

		}
		// This is where you do the 
		if ( strnicmp(requestheaderP->filepathname,wwwroot,strlen(wwwroot)) == 0 )  // else security violation!
		{

		}
		else
		{
			OutputHTTPError(client_socket, 403);	// 403 - forbidden
			return;
		}
	}
	else
	{
		OutputHTTPError(client_socket, 501);   // 501 not implemented
		return;
	}

}


//--------------------------------------------------------------
//	SocketRead()
//		Reads data from the client socket until it gets a valid http
//		header or the client disconnects.
//--------------------------------------------------------------
int SocketRead(SOCKET client_socket, char *receivebuffer, int buffersize)
{
	int size = 0, totalsize = 0;

	do
	{
		size = recv(client_socket,receivebuffer+totalsize,buffersize-totalsize,0);
		if ( size != 0 && size != SOCKET_ERROR )
		{
			totalsize += size;

			// are we done reading the http header?
			if ( strstr(receivebuffer,"\r\n\r\n") )
				break;
		}
		else
			totalsize = size;			// remember error state for return
		
	} while ( size != 0 && size != SOCKET_ERROR );

	return(totalsize);
}


//--------------------------------------------------------------
//	OutputScreenError()
//		Writes an error message to the screen displays the socket
//		error code, clearing the error before exiting.
//--------------------------------------------------------------
void OutputScreenError(const char *errmsg)
{
//	EnterCriticalSection (&output_criticalsection);
//   cout << errmsg << " - " << WSAGetLastError() << endl;
//   WSASetLastError(0);
//	LeaveCriticalSection (&output_criticalsection);
}


//--------------------------------------------------------------
//	OutputHTTPError()
//		Sends an http header and html body to the client with
//		error information.
//--------------------------------------------------------------
void OutputHTTPError(SOCKET client_socket, int statuscode)
{
	char headerbuffer[COMM_BUFFER_SIZE];
	char htmlbuffer[COMM_BUFFER_SIZE];

	sprintf(htmlbuffer,"<html><body><h2>Error: %d</h2></body></html>",statuscode);
	sprintf(headerbuffer,"HTTP/1.0 %d\r\nContent-Type: text/html\r\nContent-Length: %ld\r\n\r\n",statuscode,strlen(htmlbuffer));
	
	send(client_socket,headerbuffer,strlen(headerbuffer),0);
	send(client_socket,htmlbuffer,strlen(htmlbuffer),0);

	closesocket(client_socket);
}


//--------------------------------------------------------------
//	OutputHTTPRedirect()
//		Writes an HTTP redirect header and body to the client.
//		Called if the user requests a directory causing the redirect
//		to directory/index.html 
//--------------------------------------------------------------
void OutputHTTPRedirect(SOCKET client_socket, const char *defaulturl)
{
	char headerbuffer[COMM_BUFFER_SIZE];
	char htmlbuffer[COMM_BUFFER_SIZE];
	char hosturl[MAX_PATH];

	sprintf(hosturl,"http://%s",hostname);
	strncat(hosturl,defaulturl,COMM_BUFFER_SIZE);

	if ( hosturl[strlen(hosturl)-1] != '/' )
		strncat(hosturl,"/",MAX_PATH);										
	strncat(hosturl,"index.html",MAX_PATH);

	sprintf(htmlbuffer,"<html><body><a href=\"%s\">%s</a></body></html>",hosturl,hosturl);
	sprintf(headerbuffer,"HTTP/1.0 301\r\nContent-Type: text/html\r\nContent-Length: %ld\r\nLocation: %s\r\n\r\n",strlen(htmlbuffer),hosturl);
	
	send(client_socket,headerbuffer,strlen(headerbuffer),0);
	send(client_socket,htmlbuffer,strlen(htmlbuffer),0);

	closesocket(client_socket);
}

//--------------------------------------------------------------
//	ParseHTTPHeader()
//		Fills a HTTPRequestHeader with method, url, http version
//		and file system path information.
//--------------------------------------------------------------
BOOL ParseHTTPHeader(char *receivebuffer, HTTPRequestHeader *requestheaderP)
{
	char *pos;
	// http request header format
	// method uri httpversion

	//debugging
	//EnterCriticalSection (&output_criticalsection);
   //cout << receivebuffer << endl;
	//LeaveCriticalSection (&output_criticalsection);
	// end debuggine	



	pos = strtok(receivebuffer," ");
	if ( pos == NULL )
		return(FALSE);
	strncpy(requestheaderP->method,pos,SMALL_BUFFER_SIZE);
	
	pos = strtok(NULL," ");
	if ( pos == NULL )
		return(FALSE);
	strncpy(requestheaderP->url,pos,MAX_PATH);
	decodeURLEncode(requestheaderP->url);

	pos = strtok(NULL,"\r");
	if ( pos == NULL )
		return(FALSE);
	strncpy(requestheaderP->httpversion,pos,SMALL_BUFFER_SIZE);

	// based on the url lets figure out the filename + path
	strncpy(requestheaderP->filepathname,wwwroot,MAX_PATH);
	strncat(requestheaderP->filepathname,requestheaderP->url,MAX_PATH);

	// because the filepathname can have relative references  ../ ./ 
	// call _fullpath to get the absolute 'real' filepath
	// _fullpath seems to handle '/' and '\' 
   _fullpath(requestheaderP->filepathname,requestheaderP->filepathname,MAX_PATH);

	return(TRUE);
}


//--------------------------------------------------------------
//	StartWebServer()
//		Creates server sock and binds to ip address and port
//--------------------------------------------------------------
SOCKET StartWebServer()
{
	SOCKET s;

	s = socket(AF_INET,SOCK_STREAM,0);
	if ( s == INVALID_SOCKET )
	{
		OutputScreenError("Error creating sock()");
		return(0);
	}
	
	SOCKADDR_IN si;

	si.sin_family = AF_INET;
	si.sin_port = htons(wwwport);		// port
	si.sin_addr.s_addr = htonl(INADDR_ANY);

	if ( bind(s,(struct sockaddr *) &si,sizeof(SOCKADDR_IN)) == SOCKET_ERROR )
	{
		OutputScreenError("Error in bind()");
		closesocket(s);
		return(0);
	}

	return(s);
}

//--------------------------------------------------------------
//	findMimeType()
//		Performs linear search through mimetypes array looking for
//		matching file extension returning index of mime type 
//--------------------------------------------------------------
int findMimeType(const char *filename)
{
	char *pos;
	int numofelements;
	
	pos = strrchr(filename,'.');

	if ( pos )
	{
		numofelements = sizeof(mimetypes) / sizeof(MimeAssociation);

		for ( int x = 0; x < numofelements; ++x )
		{
			if ( stricmp(mimetypes[x].file_ext,pos) == 0 )
				return(x);
		}
	}

	return(0);  // return default mimetype  'text/plain' 
}

//--------------------------------------------------------------
//	DetermineHost()
//		If webserver needs to redirect user from directory to 
//		default html file the server builds a full url and hence
//		needs it's full domain name for http address.
//			http://mymachine.rollins.brevard.edu/index.html
//--------------------------------------------------------------
void DetermineHost( char *hostname )
{
	IN_ADDR in;
   hostent *h;

   gethostname(hostname,MAX_PATH);
   h = gethostbyname(hostname);
	memcpy(&in,h->h_addr,4);
	h = gethostbyaddr((char *)&in,4,PF_INET);
	strcpy(hostname,h->h_name);
}
